# Needed on case-insensitive filesystems
from __future__ import absolute_import

# Try to import PIL in either of the two ways it can be installed.
try:
    from PIL import Image, ImageDraw
except ImportError:  # pragma: no cover
    import Image
    import ImageDraw

import qrcode.image.base

from qrcode.image.styles.colormasks import SolidFillColorMask
from qrcode.image.styles.moduledrawers import SquareModuleDrawer
class StyledPilImage(qrcode.image.base.BaseImage):
    """
    Styled PIL image builder, default format is PNG.
    This differs from the PilImage in that there is a module_drawer, a color_mask, and an optional image

    The module_drawer should extend the QRModuleDrawer class and implement the
    drawrect_context(self, box, active, context), and probably also the intitialize
    function. This will draw an individual "module" or square on the QR code.

    The color_mask will extend the QRColorMask class and will at very least implement the
    get_fg_pixel(image, x, y) function, calculating a color to put on the image at the pixel location (x,y)
    (more advanced functionality can be gotten by instead overriding other functions defined in the QRColorMask class)


    The Image can be specified either by path or with a Pillow Image, and if it is there will be placed in the
    middle of the QR code. No effort is done to ensure that the QR code is still legible after the image has been
    placed there; Q or H level error correction levels are recommended to maintain data integrity
    A resampling filter can be specified (defaulting to PIL.Image.LANCZOS) for resizing; see PIL.Image.resize() for
    possible options for this parameter.
    """
    kind = "PNG"

    needs_context = True
    needs_processing = True

    def new_image(self, **kwargs):
        self.color_mask = kwargs.get("color_mask", SolidFillColorMask())
        self.module_drawer = kwargs.get("module_drawer", SquareModuleDrawer())
        # The eye drawer can be overridden by another module drawer as well, but you have to be more careful with these in order to
        # make the QR code still parseable
        self.eye_drawer = kwargs.get("eye_drawer", SquareModuleDrawer())

        embeded_image_path = kwargs.get("embeded_image_path", None)
        self.embeded_image = kwargs.get("embeded_image", None)
        self.embeded_image_resample = kwargs.get("embeded_image_resample", Image.LANCZOS)
        if not self.embeded_image and embeded_image_path:
            self.embeded_image = Image.open(embeded_image_path)
        mode = "RGBA" if (self.color_mask.has_transparency or (self.embeded_image and 'A' in self.embeded_image.getbands())) else "RGB"
        self.mode = mode

        self.back_color = self.color_mask.back_color # This is the background color. Should be white or whiteish

        img = Image.new(mode, (self.pixel_size, self.pixel_size), self.back_color)

        # the paint_color is the color the module drawer will use to draw upon a canvas
        # During the color mask process, pixels that are paint_color are replaced by a newly-calculated color
        self.paint_color = tuple(0 for i in self.color_mask.back_color)
        if self.color_mask.has_transparency:
            self.paint_color = tuple([*self.color_mask.back_color[:3], 255])

        self.color_mask.initialize(self, img)
        self.module_drawer.initialize(self, img)
        self.eye_drawer.initialize(self, img)
        return img

    def drawrect_context(self, row, col, is_active, context):
        box = self.pixel_box(row, col)
        if self.is_eye(row, col):
            self.eye_drawer.drawrect_context(box, is_active, context)
        else:
            self.module_drawer.drawrect_context(box, is_active, context)

    def process(self):
        self.color_mask.apply_mask(self._img)
        if self.embeded_image:
            self.draw_embeded_image()

    def draw_embeded_image(self):
            total_width,_ = self._img.size
            total_width = int(total_width)
            logo_width_ish = int(total_width / 4)
            logo_offset = int( (int(total_width / 2) - int(logo_width_ish / 2)) / self.box_size) * self.box_size # round the offset to the nearest module
            logo_position = (logo_offset, logo_offset)
            logo_width = total_width - logo_offset*2
            region = self.embeded_image
            region = region.resize((logo_width, logo_width), self.embeded_image_resample)
            if 'A' in region.getbands():
                self._img.alpha_composite(region, logo_position)
            else:
                self._img.paste(region, logo_position)

    # The eyes are treated differently, and this will find whether the referenced module is in an eye
    def is_eye(self, row, col):
        return (
            (row < 7 and col < 7)
            or (row < 7 and self.width - col < 8)
            or (self.width - row < 8 and col < 7)
        )

    def save(self, stream, format=None, **kwargs):
        if format is None:
            format = kwargs.get("kind", self.kind)
        if "kind" in kwargs:
            del kwargs["kind"]
        self._img.save(stream, format=format, **kwargs)

    def __getattr__(self, name):
        return getattr(self._img, name)



# Needed on case-insensitive filesystems
from __future__ import absolute_import

# Try to import PIL in either of the two ways it can be installed.
try:
    from PIL import Image, ImageDraw
except ImportError:  # pragma: no cover
    import Image
    import ImageDraw

# When drawing antialiased things, make them bigger and then shrink them down to size after the geometry has been drawn
ANTIALIASING_FACTOR = 4

class QRModuleDrawer:
    """
    QRModuleDrawer exists to draw the modules of the  QR Code onto a PIL image
    For this, technically all that is necessary is a drawrect_context(self, box, is_active, context) function
    which takes in the box in which it is to draw, whether or not the box is "active" (a module exists there),
    and the context (the neighboring pixels)
    It is frequently necessary to also implement an "initialize" function to set up values that only the containing StyledPilImage
    knows about.
    NOTE: the color that this draws in should be whatever is equivalent to black in the color space, and the specified QRColorMask
    will handle adding colors as necessary to the image

    For examples of what these look like, see doc/module_drawers.png
    """

    fill = None
    def initialize(self, styledPilImage, image):
        self.fill = styledPilImage.paint_color

    def drawrect_context(self, box, is_active, context):
        raise NotImplementedError("QRModuleDrawer.drawrect_context")

    # helper for figuring out the context, which is an array containing information on neighboring pixels
    # I refer to these by their cardinal directions, like:
    # [NW, N,  NE,
    #  W,      E,
    #  SW, S,  SE]
    DIRECTIONS = {
        'NW': 0,
        'N': 1,
        'NE': 2,
        'W': 3,
        'E': 4,
        'SW': 5,
        'S': 6,
        'SE': 7
    }
    def get(self, context, direction):
        return context[self.DIRECTIONS[direction]]

class SquareModuleDrawer(QRModuleDrawer):
    """
    Draws the modules as simple squares
    """
    fill = None
    def initialize(self, styledPilImage, image):
        self.imgDraw = ImageDraw.Draw(image)
        self.fill = styledPilImage.paint_color
    def drawrect_context(self, box, is_active, context):
        if is_active:
            self.imgDraw.rectangle(box, fill=self.fill)

class GappedSquareModuleDrawer(QRModuleDrawer):
    """
    Draws the modules as simple squares that are not contiguous.
    The size_ratio determines how wide the squares are relative to the width of the space they are printed in
    """
    fill = None
    def __init__(self, size_ratio = 0.8):
        self.size_ratio = size_ratio
    def initialize(self, styledPilImage, image):
        self.imgDraw = ImageDraw.Draw(image)
        self.fill = styledPilImage.paint_color
        self.delta = (1 - self.size_ratio) * styledPilImage.box_size / 2
    def drawrect_context(self, box, is_active, context):
        if is_active:
            smaller_box = (
                box[0][0] + self.delta,
                box[0][1] + self.delta,
                box[1][0] - self.delta,
                box[1][1] - self.delta
            )
            self.imgDraw.rectangle(smaller_box, fill=self.fill)

class CircleModuleDrawer(QRModuleDrawer):
    """
    Draws the modules as circles
    """
    circle = None
    def initialize(self, styledPilImage, image):
        box_size = styledPilImage.box_size
        fake_size = box_size * ANTIALIASING_FACTOR
        self.circle = Image.new(styledPilImage.mode,(fake_size, fake_size), styledPilImage.color_mask.back_color)
        ImageDraw.Draw(self.circle).ellipse((0,0, fake_size, fake_size), fill = styledPilImage.paint_color)
        self.circle = self.circle.resize((box_size, box_size), Image.LANCZOS)
        self.image = image
    def drawrect_context(self, box, is_active, context):
        if is_active:
            self.image.paste(self.circle, (box[0][0], box[0][1]))

class RoundedModuleDrawer(QRModuleDrawer):
    """
    Draws the modules with all 90 degree corners replaced with rounded edges
    radius_ratio determines the radius of the rounded edges -
    a value of 1 means that an isolated module will be drawn as a circle,
    while a value of 0 means that the radius of the rounded edge will be 0 (and thus back to 90 degrees again)
    """
    def __init__(self, radius_ratio = 1):
        self.radius_ratio = radius_ratio

    def initialize(self, styledPilImage, image):
        self.corner_width = int(styledPilImage.box_size / 2)
        self.image = image
        self.setup_corners(styledPilImage.mode, styledPilImage.color_mask.back_color, styledPilImage.paint_color)
    def setup_corners(self, mode, back_color, front_color):
        self.SQUARE = Image.new(mode, (self.corner_width, self.corner_width), front_color)

        fake_width = self.corner_width * ANTIALIASING_FACTOR
        radius = self.radius_ratio * fake_width
        diameter = radius * 2
        base = Image.new(mode, (fake_width, fake_width), back_color) # make something 4x bigger for antialiasing
        base_draw = ImageDraw.Draw(base)
        base_draw.ellipse((0,0, diameter, diameter), fill = front_color)
        base_draw.rectangle((radius, 0, fake_width, fake_width), fill = front_color)
        base_draw.rectangle((0, radius, fake_width, fake_width), fill = front_color)
        self.NW_ROUND = base.resize((self.corner_width, self.corner_width), Image.LANCZOS)
        self.SW_ROUND = self.NW_ROUND.transpose(Image.FLIP_TOP_BOTTOM)
        self.SE_ROUND = self.NW_ROUND.transpose(Image.ROTATE_180)
        self.NE_ROUND = self.NW_ROUND.transpose(Image.FLIP_LEFT_RIGHT)

    def drawrect_context(self, box, is_active, context):
        if not is_active:
            return
        # find rounded edges
        nw_rounded = not self.get(context, 'W') and not self.get(context, 'N')
        ne_rounded = not self.get(context, 'N') and not self.get(context, 'E')
        se_rounded = not self.get(context, 'E') and not self.get(context, 'S')
        sw_rounded = not self.get(context, 'S') and not self.get(context, 'W')

        nw = self.NW_ROUND if nw_rounded else self.SQUARE
        ne = self.NE_ROUND if ne_rounded else self.SQUARE
        se = self.SE_ROUND if se_rounded else self.SQUARE
        sw = self.SW_ROUND if sw_rounded else self.SQUARE
        self.image.paste(nw, (box[0][0], box[0][1]))
        self.image.paste(ne, (box[0][0] + self.corner_width, box[0][1]))
        self.image.paste(se, (box[0][0] + self.corner_width, box[0][1] + self.corner_width))
        self.image.paste(sw, (box[0][0], box[0][1] + self.corner_width))


class VerticalBarsDrawer(QRModuleDrawer):
    """
    Draws vertically contiguous groups of modules as long rounded rectangles, with gaps between neighboring bands
    (the size of these gaps is inversely proportional to the horizontal_shrink)
    """
    def __init__(self, horizontal_shrink = 0.8):
        self.horizontal_shrink = horizontal_shrink

    def initialize(self, styledPilImage, image):
        self.half_height = int(styledPilImage.box_size / 2)
        self.image = image
        self.delta = int((1-self.horizontal_shrink) * self.half_height)
        self.setup_edges(styledPilImage.mode, styledPilImage.color_mask.back_color, styledPilImage.paint_color)

    def setup_edges(self, mode, back_color, front_color):
        height = self.half_height
        width = height * 2
        shrunken_width = int(width * self.horizontal_shrink)
        self.SQUARE = Image.new(mode, (shrunken_width, height), front_color)

        fake_width = width * ANTIALIASING_FACTOR
        fake_height = height * ANTIALIASING_FACTOR
        radius = fake_width
        base = Image.new(mode, (fake_width, fake_height), back_color) # make something 4x bigger for antialiasing
        base_draw = ImageDraw.Draw(base)
        base_draw.ellipse((0,0, fake_width, fake_height * 2), fill = front_color)

        self.ROUND_TOP = base.resize((shrunken_width, height), Image.LANCZOS)
        self.ROUND_BOTTOM = self.ROUND_TOP.transpose(Image.FLIP_TOP_BOTTOM)

    def drawrect_context(self, box, is_active, context):
        if is_active:
            # find rounded edges
            top_rounded = not self.get(context, 'N')
            bottom_rounded = not self.get(context, 'S')

            top = self.ROUND_TOP if top_rounded else self.SQUARE
            bottom = self.ROUND_BOTTOM if bottom_rounded else self.SQUARE
            self.image.paste(top, (box[0][0] + self.delta, box[0][1]))
            self.image.paste(bottom, (box[0][0] + self.delta, box[0][1] + self.half_height))

class HorizontalBarsDrawer(QRModuleDrawer):
    """
    Draws horizontally contiguous groups of modules as long rounded rectangles, with gaps between neighboring bands
    (the size of these gaps is inversely proportional to the vertical_shrink)
    """
    def __init__(self, vertical_shrink = 0.8):
        self.vertical_shrink = vertical_shrink

    def initialize(self, styledPilImage, image):
        self.half_width = int(styledPilImage.box_size / 2)
        self.image = image
        self.delta = int((1-self.vertical_shrink) * self.half_width)
        self.setup_edges(styledPilImage.mode, styledPilImage.color_mask.back_color, styledPilImage.paint_color)

    def setup_edges(self, mode, back_color, front_color):
        width = self.half_width
        height = width * 2
        shrunken_height= int(height * self.vertical_shrink)
        self.SQUARE = Image.new(mode, (width, shrunken_height), front_color)

        fake_width = width * ANTIALIASING_FACTOR
        fake_height = height * ANTIALIASING_FACTOR
        radius = fake_height
        base = Image.new(mode, (fake_width, fake_height), back_color) # make something 4x bigger for antialiasing
        base_draw = ImageDraw.Draw(base)
        base_draw.ellipse((0,0, fake_width * 2, fake_height), fill = front_color)

        self.ROUND_LEFT = base.resize((width, shrunken_height), Image.LANCZOS)
        self.ROUND_RIGHT = self.ROUND_LEFT.transpose(Image.FLIP_LEFT_RIGHT)

    def drawrect_context(self, box, is_active, context):
        if is_active:
            # find rounded edges
            left_rounded = not self.get(context, 'W')
            right_rounded = not self.get(context, 'E')

            left = self.ROUND_LEFT if left_rounded else self.SQUARE
            right = self.ROUND_RIGHT if right_rounded else self.SQUARE
            self.image.paste(left, (box[0][0], box[0][1] + self.delta))
            self.image.paste(right, (box[0][0] + self.half_width, box[0][1] + self.delta))



import io
import os
import sys
import unittest
import warnings
from tempfile import mkdtemp
from unittest import mock

try:
    import pymaging_png  # ensure that PNG support is installed
    import qrcode.image.pure
except ImportError:  # pragma: no cover
    pymaging_png = None

import qrcode
import qrcode.image.svg
import qrcode.util
from qrcode.exceptions import DataOverflowError
from qrcode.image.base import BaseImage
from qrcode.image.pil import Image as pil_Image
from qrcode.image.styledpil import StyledPilImage
from qrcode.image.styles import colormasks, moduledrawers
from qrcode.tests.svg import SvgImageWhite
from qrcode.util import MODE_8BIT_BYTE, MODE_ALPHA_NUM, MODE_NUMBER, QRData

UNICODE_TEXT = '\u03b1\u03b2\u03b3'
WHITE = (255, 255, 255)
BLACK = (0, 0, 0)
RED = (255, 0, 0)


class QRCodeTests(unittest.TestCase):
    def setUp(self):
        self.tmpdir = mkdtemp()

    def tearDown(self):
        os.rmdir(self.tmpdir)

    def test_basic(self):
        qr = qrcode.QRCode(version=1)
        qr.add_data('a')
        qr.make(fit=False)

    def test_large(self):
        qr = qrcode.QRCode(version=27)
        qr.add_data('a')
        qr.make(fit=False)

    def test_invalid_version(self):
        qr = qrcode.QRCode(version=41)
        self.assertRaises(ValueError, qr.make, fit=False)

    def test_invalid_border(self):
        self.assertRaises(ValueError, qrcode.QRCode, border=-1)

    def test_overflow(self):
        qr = qrcode.QRCode(version=1)
        qr.add_data('abcdefghijklmno')
        self.assertRaises(DataOverflowError, qr.make, fit=False)

    def test_add_qrdata(self):
        qr = qrcode.QRCode(version=1)
        data = QRData('a')
        qr.add_data(data)
        qr.make(fit=False)

    def test_fit(self):
        qr = qrcode.QRCode()
        qr.add_data('a')
        qr.make()
        self.assertEqual(qr.version, 1)
        qr.add_data('bcdefghijklmno')
        qr.make()
        self.assertEqual(qr.version, 2)

    def test_mode_number(self):
        qr = qrcode.QRCode()
        qr.add_data('1234567890123456789012345678901234', optimize=0)
        qr.make()
        self.assertEqual(qr.version, 1)
        self.assertEqual(qr.data_list[0].mode, MODE_NUMBER)

    def test_mode_alpha(self):
        qr = qrcode.QRCode()
        qr.add_data('ABCDEFGHIJ1234567890', optimize=0)
        qr.make()
        self.assertEqual(qr.version, 1)
        self.assertEqual(qr.data_list[0].mode, MODE_ALPHA_NUM)

    def test_regression_mode_comma(self):
        qr = qrcode.QRCode()
        qr.add_data(',', optimize=0)
        qr.make()
        self.assertEqual(qr.data_list[0].mode, MODE_8BIT_BYTE)

    def test_mode_8bit(self):
        qr = qrcode.QRCode()
        qr.add_data('abcABC' + UNICODE_TEXT, optimize=0)
        qr.make()
        self.assertEqual(qr.version, 1)
        self.assertEqual(qr.data_list[0].mode, MODE_8BIT_BYTE)

    def test_mode_8bit_newline(self):
        qr = qrcode.QRCode()
        qr.add_data('ABCDEFGHIJ1234567890\n', optimize=0)
        qr.make()
        self.assertEqual(qr.data_list[0].mode, MODE_8BIT_BYTE)

    def test_render_pil(self):
        qr = qrcode.QRCode()
        qr.add_data(UNICODE_TEXT)
        img = qr.make_image()
        img.save(io.BytesIO())
        self.assertIsInstance(img.get_image(), pil_Image.Image)

    def test_render_pil_with_transparent_background(self):
        qr = qrcode.QRCode()
        qr.add_data(UNICODE_TEXT)
        img = qr.make_image(back_color='TransParent')
        img.save(io.BytesIO())

    def test_render_pil_with_red_background(self):
        qr = qrcode.QRCode()
        qr.add_data(UNICODE_TEXT)
        img = qr.make_image(back_color='red')
        img.save(io.BytesIO())

    def test_render_pil_with_rgb_color_tuples(self):
        qr = qrcode.QRCode()
        qr.add_data(UNICODE_TEXT)
        img = qr.make_image(back_color=(255, 195, 235), fill_color=(55, 95, 35))
        img.save(io.BytesIO())

    def test_render_with_pattern(self):
        qr = qrcode.QRCode(mask_pattern=3)
        qr.add_data(UNICODE_TEXT)
        img = qr.make_image()
        img.save(io.BytesIO())

    def test_make_image_with_wrong_pattern(self):
        with self.assertRaises(TypeError):
            qrcode.QRCode(mask_pattern='string pattern')

        with self.assertRaises(ValueError):
            qrcode.QRCode(mask_pattern=-1)

        with self.assertRaises(ValueError):
            qrcode.QRCode(mask_pattern=42)

    def test_mask_pattern_setter(self):
        qr = qrcode.QRCode()

        with self.assertRaises(TypeError):
            qr.mask_pattern = "string pattern"

        with self.assertRaises(ValueError):
            qr.mask_pattern = -1

        with self.assertRaises(ValueError):
            qr.mask_pattern = 8

    def test_qrcode_bad_factory(self):
        with self.assertRaises(TypeError):
           qrcode.QRCode(image_factory='not_BaseImage')

        with self.assertRaises(AssertionError):
            qrcode.QRCode(image_factory=dict)

    def test_qrcode_factory(self):

        class MockFactory(BaseImage):
            drawrect = mock.Mock()
            new_image = mock.Mock()

        qr = qrcode.QRCode(image_factory=MockFactory)
        qr.add_data(UNICODE_TEXT)
        qr.make_image()
        self.assertTrue(MockFactory.new_image.called)
        self.assertTrue(MockFactory.drawrect.called)

    def test_render_svg(self):
        qr = qrcode.QRCode()
        qr.add_data(UNICODE_TEXT)
        img = qr.make_image(image_factory=qrcode.image.svg.SvgImage)
        img.save(io.BytesIO())

    def test_render_svg_path(self):
        qr = qrcode.QRCode()
        qr.add_data(UNICODE_TEXT)
        img = qr.make_image(image_factory=qrcode.image.svg.SvgPathImage)
        img.save(io.BytesIO())

    def test_render_svg_fragment(self):
        qr = qrcode.QRCode()
        qr.add_data(UNICODE_TEXT)
        img = qr.make_image(image_factory=qrcode.image.svg.SvgFragmentImage)
        img.save(io.BytesIO())

    def test_svg_string(self):
        qr = qrcode.QRCode()
        qr.add_data(UNICODE_TEXT)
        img = qr.make_image(image_factory=qrcode.image.svg.SvgFragmentImage)
        file_like = io.BytesIO()
        img.save(file_like)
        file_like.seek(0)
        assert file_like.read() in img.to_string()

    def test_render_svg_with_background(self):
        qr = qrcode.QRCode()
        qr.add_data(UNICODE_TEXT)
        img = qr.make_image(image_factory=SvgImageWhite)
        img.save(io.BytesIO())

    @unittest.skipIf(not pymaging_png, "Requires pymaging with PNG support")
    def test_render_pymaging_png(self):
        qr = qrcode.QRCode()
        qr.add_data(UNICODE_TEXT)
        img = qr.make_image(image_factory=qrcode.image.pure.PymagingImage)
        from pymaging import Image as pymaging_Image
        self.assertIsInstance(img.get_image(), pymaging_Image)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', DeprecationWarning)
            img.save(io.BytesIO())

    @unittest.skipIf(not pymaging_png, "Requires pymaging")
    def test_render_pymaging_png_bad_kind(self):
        qr = qrcode.QRCode()
        qr.add_data(UNICODE_TEXT)
        img = qr.make_image(image_factory=qrcode.image.pure.PymagingImage)
        with self.assertRaises(ValueError):
            img.save(io.BytesIO(), kind='FISH')

    def test_render_styled_pil_image(self):
        qr = qrcode.QRCode(error_correction=qrcode.constants.ERROR_CORRECT_L)
        qr.add_data(UNICODE_TEXT)
        img = qr.make_image(image_factory=StyledPilImage)
        img.save(io.BytesIO())

    def test_render_styled_with_embeded_image(self):
        embeded_img = pil_Image.new('RGB', (10, 10), color='red')
        qr = qrcode.QRCode(error_correction=qrcode.constants.ERROR_CORRECT_L)
        qr.add_data(UNICODE_TEXT)
        img = qr.make_image(image_factory=StyledPilImage, embeded_image=embeded_img)
        img.save(io.BytesIO())

    def test_render_styled_with_embeded_image_path(self):
        tmpfile = os.path.join(self.tmpdir, "test.png")
        embeded_img = pil_Image.new('RGB', (10, 10), color='red')
        embeded_img.save(tmpfile)
        qr = qrcode.QRCode(error_correction=qrcode.constants.ERROR_CORRECT_L)
        qr.add_data(UNICODE_TEXT)
        img = qr.make_image(image_factory=StyledPilImage, embeded_image_path=tmpfile)
        img.save(io.BytesIO())
        os.remove(tmpfile)

    def test_render_styled_with_square_module_drawer(self):
        qr = qrcode.QRCode(error_correction=qrcode.constants.ERROR_CORRECT_L)
        qr.add_data(UNICODE_TEXT)
        img = qr.make_image(image_factory=StyledPilImage, module_drawer=moduledrawers.SquareModuleDrawer())
        img.save(io.BytesIO())

    def test_render_styled_with_gapped_module_drawer(self):
        qr = qrcode.QRCode(error_correction=qrcode.constants.ERROR_CORRECT_L)
        qr.add_data(UNICODE_TEXT)
        img = qr.make_image(image_factory=StyledPilImage, module_drawer=moduledrawers.GappedSquareModuleDrawer())
        img.save(io.BytesIO())

    def test_render_styled_with_circle_module_drawer(self):
        qr = qrcode.QRCode(error_correction=qrcode.constants.ERROR_CORRECT_L)
        qr.add_data(UNICODE_TEXT)
        img = qr.make_image(image_factory=StyledPilImage, module_drawer=moduledrawers.CircleModuleDrawer())
        img.save(io.BytesIO())

    def test_render_styled_with_rounded_module_drawer(self):
        qr = qrcode.QRCode(error_correction=qrcode.constants.ERROR_CORRECT_L)
        qr.add_data(UNICODE_TEXT)
        img = qr.make_image(image_factory=StyledPilImage, module_drawer=moduledrawers.RoundedModuleDrawer())
        img.save(io.BytesIO())

    def test_render_styled_with_vertical_bars_module_drawer(self):
        qr = qrcode.QRCode(error_correction=qrcode.constants.ERROR_CORRECT_L)
        qr.add_data(UNICODE_TEXT)
        img = qr.make_image(image_factory=StyledPilImage, module_drawer=moduledrawers.VerticalBarsDrawer())
        img.save(io.BytesIO())

    def test_render_styled_with_horizontal_bars_module_drawer(self):
        qr = qrcode.QRCode(error_correction=qrcode.constants.ERROR_CORRECT_L)
        qr.add_data(UNICODE_TEXT)
        img = qr.make_image(image_factory=StyledPilImage, module_drawer=moduledrawers.HorizontalBarsDrawer())
        img.save(io.BytesIO())

    def test_render_styled_with_default_solid_color_mask(self):
        qr = qrcode.QRCode(error_correction=qrcode.constants.ERROR_CORRECT_L)
        qr.add_data(UNICODE_TEXT)
        mask = colormasks.SolidFillColorMask()
        img = qr.make_image(image_factory=StyledPilImage, color_mask=mask)
        img.save(io.BytesIO())

    def test_render_styled_with_solid_color_mask(self):
        qr = qrcode.QRCode(error_correction=qrcode.constants.ERROR_CORRECT_L)
        qr.add_data(UNICODE_TEXT)
        mask = colormasks.SolidFillColorMask(back_color=WHITE, front_color=RED)
        img = qr.make_image(image_factory=StyledPilImage, color_mask=mask)
        img.save(io.BytesIO())

    def test_render_styled_with_color_mask_with_transparency(self):
        qr = qrcode.QRCode(error_correction=qrcode.constants.ERROR_CORRECT_L)
        qr.add_data(UNICODE_TEXT)
        mask = colormasks.SolidFillColorMask(back_color=(255, 0, 255, 255), front_color=RED)
        img = qr.make_image(image_factory=StyledPilImage, color_mask=mask)
        img.save(io.BytesIO())
        assert img.mode == "RGBA"

    def test_render_styled_with_radial_gradient_color_mask(self):
        qr = qrcode.QRCode(error_correction=qrcode.constants.ERROR_CORRECT_L)
        qr.add_data(UNICODE_TEXT)
        mask = colormasks.RadialGradiantColorMask(back_color=WHITE, center_color=BLACK, edge_color=RED)
        img = qr.make_image(image_factory=StyledPilImage, color_mask=mask)
        img.save(io.BytesIO())

    def test_render_styled_with_square_gradient_color_mask(self):
        qr = qrcode.QRCode(error_correction=qrcode.constants.ERROR_CORRECT_L)
        qr.add_data(UNICODE_TEXT)
        mask = colormasks.SquareGradiantColorMask(back_color=WHITE, center_color=BLACK, edge_color=RED)
        img = qr.make_image(image_factory=StyledPilImage, color_mask=mask)
        img.save(io.BytesIO())

    def test_render_styled_with_horizontal_gradient_color_mask(self):
        qr = qrcode.QRCode(error_correction=qrcode.constants.ERROR_CORRECT_L)
        qr.add_data(UNICODE_TEXT)
        mask = colormasks.HorizontalGradiantColorMask(back_color=WHITE, left_color=RED, right_color=BLACK)
        img = qr.make_image(image_factory=StyledPilImage, color_mask=mask)
        img.save(io.BytesIO())

    def test_render_styled_with_vertical_gradient_color_mask(self):
        qr = qrcode.QRCode(error_correction=qrcode.constants.ERROR_CORRECT_L)
        qr.add_data(UNICODE_TEXT)
        mask = colormasks.VerticalGradiantColorMask(back_color=WHITE, top_color=RED, bottom_color=BLACK)
        img = qr.make_image(image_factory=StyledPilImage, color_mask=mask)
        img.save(io.BytesIO())

    def test_render_styled_with_image_color_mask(self):
        img_mask = pil_Image.new('RGB', (10, 10), color='red')
        qr = qrcode.QRCode(error_correction=qrcode.constants.ERROR_CORRECT_L)
        qr.add_data(UNICODE_TEXT)
        mask = colormasks.ImageColorMask(back_color=WHITE, color_mask_image=img_mask)
        img = qr.make_image(image_factory=StyledPilImage, color_mask=mask)
        img.save(io.BytesIO())

    def test_optimize(self):
        qr = qrcode.QRCode()
        text = 'A1abc12345def1HELLOa'
        qr.add_data(text, optimize=4)
        qr.make()
        self.assertEqual(
            [d.mode for d in qr.data_list],
            [
                MODE_8BIT_BYTE, MODE_NUMBER, MODE_8BIT_BYTE, MODE_ALPHA_NUM,
                MODE_8BIT_BYTE
            ]
        )
        self.assertEqual(qr.version, 2)

    def test_optimize_short(self):
        qr = qrcode.QRCode()
        text = 'A1abc1234567def1HELLOa'
        qr.add_data(text, optimize=7)
        qr.make()
        self.assertEqual(len(qr.data_list), 3)
        self.assertEqual(
            [d.mode for d in qr.data_list],
            [MODE_8BIT_BYTE, MODE_NUMBER, MODE_8BIT_BYTE]
        )
        self.assertEqual(qr.version, 2)

    def test_optimize_longer_than_data(self):
        qr = qrcode.QRCode()
        text = 'ABCDEFGHIJK'
        qr.add_data(text, optimize=12)
        self.assertEqual(len(qr.data_list), 1)
        self.assertEqual(qr.data_list[0].mode, MODE_ALPHA_NUM)

    def test_optimize_size(self):
        text = 'A1abc12345123451234512345def1HELLOHELLOHELLOHELLOa' * 5

        qr = qrcode.QRCode()
        qr.add_data(text)
        qr.make()
        self.assertEqual(qr.version, 10)

        qr = qrcode.QRCode()
        qr.add_data(text, optimize=0)
        qr.make()
        self.assertEqual(qr.version, 11)

    def test_qrdata_repr(self):
        data = b'hello'
        data_obj = qrcode.util.QRData(data)
        self.assertEqual(repr(data_obj), repr(data))

    def test_print_ascii_stdout(self):
        qr = qrcode.QRCode()
        stdout_encoding = sys.stdout.encoding
        with mock.patch('sys.stdout') as fake_stdout:
            # Python 2.6 needs sys.stdout.encoding to be a real string.
            sys.stdout.encoding = stdout_encoding
            fake_stdout.isatty.return_value = None
            self.assertRaises(OSError, qr.print_ascii, tty=True)
            self.assertTrue(fake_stdout.isatty.called)

    def test_print_ascii(self):
        qr = qrcode.QRCode(border=0)
        f = io.StringIO()
        qr.print_ascii(out=f)
        printed = f.getvalue()
        f.close()
        expected = '\u2588\u2580\u2580\u2580\u2580\u2580\u2588'
        self.assertEqual(printed[:len(expected)], expected)

        f = io.StringIO()
        f.isatty = lambda: True
        qr.print_ascii(out=f, tty=True)
        printed = f.getvalue()
        f.close()
        expected = (
            '\x1b[48;5;232m\x1b[38;5;255m' +
            '\xa0\u2584\u2584\u2584\u2584\u2584\xa0')
        self.assertEqual(printed[:len(expected)], expected)

    def test_print_tty_stdout(self):
        qr = qrcode.QRCode()
        with mock.patch('sys.stdout') as fake_stdout:
            fake_stdout.isatty.return_value = None
            self.assertRaises(OSError, qr.print_tty)
            self.assertTrue(fake_stdout.isatty.called)

    def test_print_tty(self):
        qr = qrcode.QRCode()
        f = io.StringIO()
        f.isatty = lambda: True
        qr.print_tty(out=f)
        printed = f.getvalue()
        f.close()
        BOLD_WHITE_BG = '\x1b[1;47m'
        BLACK_BG = '\x1b[40m'
        WHITE_BLOCK = BOLD_WHITE_BG + '  ' + BLACK_BG
        EOL = '\x1b[0m\n'
        expected = (
            BOLD_WHITE_BG + '  '*23 + EOL +
            WHITE_BLOCK + '  '*7 + WHITE_BLOCK)
        self.assertEqual(printed[:len(expected)], expected)

    def test_get_matrix(self):
        qr = qrcode.QRCode(border=0)
        qr.add_data('1')
        self.assertEqual(qr.get_matrix(), qr.modules)

    def test_get_matrix_border(self):
        qr = qrcode.QRCode(border=1)
        qr.add_data('1')
        matrix = [row[1:-1] for row in qr.get_matrix()[1:-1]]
        self.assertEqual(matrix, qr.modules)

    def test_negative_size_at_construction(self):
        self.assertRaises(ValueError, qrcode.QRCode, box_size=-1)

    def test_negative_size_at_usage(self):
        qr = qrcode.QRCode()
        qr.box_size = -1
        self.assertRaises(ValueError, qr.make_image)


class ShortcutTest(unittest.TestCase):

    def runTest(self):
        qrcode.make('image')
